from io import BytesIO
from unittest import TestCase

from dark.reads import DNARead
from dark.summarize import sequenceCategoryLengths, summarizeReads


class TestSummarizeReads(TestCase):
    """
    Test the summarizeReads function.
    """

    def testReadNumberEmptyInput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["read_number"], 0)

    def testReadNumberOneSequenceCount(self):
        seq = b">hey\nagtcagtcagtc"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["read_number"], 1)

    def testReadNumberTwoSequencesCount(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["read_number"], 2)

    def testTotalLengthEmptyInput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["total_length"], 0)

    def testTotalLengthOneString(self):
        seq = b">hey\nagtcagtcagtc"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["total_length"], 12)

    def testTotalLengthTwoStrings(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["total_length"], 17)

    def testBaseCountsEmptyImput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["base_counts"], {})

    def testBaseCountsOneRead(self):
        seq = b">hey\nagtcagtcagtc"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["base_counts"], {"a": 3, "c": 3, "t": 3, "g": 3})

    def testBaseCountsTwoReads(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["base_counts"], {"a": 4, "c": 5, "t": 4, "g": 4})

    def testMaxLengthListEmptyInput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["max_length"], 0)

    def testMaxLengthListTwoStrings(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["max_length"], 12)

    def testMinLengthListEmptyInput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["min_length"], 0)

    def testMinLengthListTwoStrings(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["min_length"], 5)

    def testMedianEmptyInput(self):
        result = summarizeReads(BytesIO(), "fasta")
        self.assertEqual(result["median_length"], 0)

    def testMedianOneString(self):
        seq = b">hey\nagtcagtcagtc"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["median_length"], 12)

    def testMedianThreeStrings(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg\n>how\natgggtc"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["median_length"], 7)

    def testMedianFourStrings(self):
        seq = b">hey\nagtcagtcagtc\n>you\nacctg\n>how\natgggtc\n>are\n\
        atggctattgaactgtatct"
        result = summarizeReads(BytesIO(seq), "fasta")
        self.assertEqual(result["median_length"], 9.5)


class TestSequenceCategoryLengths(TestCase):
    """
    Test the sequenceCategoryLengths function.
    """

    def testInvalidMinLength(self):
        """
        If a minLength value less than 1 is passed, a ValueError must be
        raised.
        """
        read = DNARead("id", "")
        error = "^minLength must be at least 1$"
        self.assertRaisesRegex(
            ValueError, error, sequenceCategoryLengths, read, {}, minLength=0
        )

    def testEmpty(self):
        """
        An empty sequence should result in an empty category summary.
        """
        read = DNARead("id", "")
        self.assertEqual([], sequenceCategoryLengths(read, {}))

    def testOneCategoryPerBase(self):
        """
        If each base is in its own category, the summary must be correct.
        """
        read = DNARead("id", "ACGT")
        categories = {
            "A": 0,
            "C": 1,
            "G": 2,
            "T": 3,
        }
        self.assertEqual(
            [(0, 1), (1, 1), (2, 1), (3, 1)], sequenceCategoryLengths(read, categories)
        )

    def testRepeatedCategory(self):
        """
        If categories are repeated in a sequence, the summary must have the
        correct length for the categories.
        """
        read = DNARead("id", "ACCGGTTT")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("a", 1), ("c", 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(read, categories),
        )

    def testUnknownCategory(self):
        """
        If a base has no category, the summary must have C{None} as the
        category for those bases.
        """
        read = DNARead("id", "ACCGGTTT")
        categories = {
            "A": "a",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("a", 1), (None, 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(read, categories),
        )

    def testUnknownCategoryWithDefault(self):
        """
        If a base has no category, the summary must have the passed default
        category as the category for those bases.
        """
        read = DNARead("id", "ACCGGTTT")
        categories = {
            "A": "a",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("a", 1), ("xxx", 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(read, categories, "xxx"),
        )

    def testSuppressAtStart(self):
        """
        If a region at the start of the sequence is shorter than the passed
        minimum length, the result should suppress the catgeory information.
        """
        read = DNARead("id", "ACCGGTTT")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("...", 1), ("c", 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(read, categories, minLength=2),
        )

    def testSuppressTwoAtStart(self):
        """
        If 2 regions at the start of the sequence are shorter than the passed
        minimum length, the result should suppress the catgeory information
        and the length of the suppressed region must be the sum of the lengths
        of the regions.
        """
        read = DNARead("id", "AGCCGGTTT")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("...", 2), ("c", 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(read, categories, minLength=2),
        )

    def testSuppressAtEnd(self):
        """
        If a region at the end of the sequence is shorter than the passed
        minimum length, the result should suppress the catgeory information.
        """
        read = DNARead("id", "CCGGTTTA")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("c", 2), ("g", 2), ("t", 3), ("...", 1)],
            sequenceCategoryLengths(read, categories, minLength=2),
        )

    def testSuppressTwoAtEnd(self):
        """
        If 2 regions at the end of the sequence are shorter than the passed
        minimum length, the result should suppress the catgeory information
        and the length of the suppressed region must be the sum of the lengths
        of the regions.
        """
        read = DNARead("id", "CCGGTTTAC")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("c", 2), ("g", 2), ("t", 3), ("...", 2)],
            sequenceCategoryLengths(read, categories, minLength=2),
        )

    def testSuppressWithNonDefaultSuppresscategory(self):
        """
        If a region of the sequence is shorter than the passed minimum length,
        the result should suppress the catgeory information and the suppress
        category returned must be the one that is passed.
        """
        read = DNARead("id", "ACCGGTTT")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("s", 1), ("c", 2), ("g", 2), ("t", 3)],
            sequenceCategoryLengths(
                read, categories, minLength=2, suppressedCategory="s"
            ),
        )

    def testAllSuppressed(self):
        """
        If all regions of the sequence are shorter than the passed
        minimum length, the result should suppress the catgeory information
        and the suppressed region length must be the sum of the region lengths.
        """
        read = DNARead("id", "ACCGGGTTT")
        categories = {
            "A": "a",
            "C": "c",
            "G": "g",
            "T": "t",
        }
        self.assertEqual(
            [("...", 9)], sequenceCategoryLengths(read, categories, minLength=5)
        )



from warnings import warn


class GenomeRanges:
    """
    Split and manage a GenBank range string as converted by BioPython.

    @param rangeStr: A C{str} indicating the (0-based) genome nucleotide
        offsets covered by a protein. This can have the following example
        forms:

        [9462:10137](+)
        [11969:12575](-)
        join{[126386:126881](-), [125941:126232](+)}
        join{[153005:153821](+), [82030:82618](-), [80480:81305](-)}

        Note that this is the string as returned from BioPython SeqIO.parse
        when it parses a GenBank flat file, not the string that is in such a
        file. For that use the C{splitRange} function you can find in an
        earlier version of this file (i.e., look in its git history).

    @raise ValueError: If C{rangeStr} cannot be correctly parsed.
    @return: A C{list} whose elements are 3-C{tuple}s holding the C{int} start
        and stop offsets and a C{bool} that indicates whether the offsets
        corresponds to the forward strand (i.e., will be C{True} when
        there is a '(+)' in a range and C{False} when there is a '(-)' which
        indicates the reverse strand).

        Example arguments and their return values:

            '[9462:10137](+)' => ((9462, 10137, True),)

            '[11969:12575](-)' => ((11969, 12575, False),)

            'join{[126386:126881](-), [125941:126232](+)}' =>
                ((126386, 126881, False),
                 (125941, 126232, True))
    """

    def __init__(self, rangeStr):
        if rangeStr.startswith("join{") and rangeStr[-1] == "}":
            join = True
            inner = rangeStr[5:-1]
        else:
            join = False
            inner = rangeStr

        ranges = []
        subRanges = inner.split(", ")
        nRanges = len(subRanges)

        if nRanges == 1 and join:
            raise ValueError(
                'Could not parse GenBank range string "%s". '
                "join{} can only be used with multiple ranges." % rangeStr
            )
        elif nRanges > 1 and not join:
            raise ValueError(
                'Could not parse GenBank range string "%s". '
                "Multiple ranges must be wrapped in join{}." % rangeStr
            )

        for subRange in subRanges:
            if subRange.endswith("](+)"):
                forward = True
            elif subRange.endswith("](-)"):
                forward = False
            else:
                raise ValueError(
                    'Could not parse GenBank range string "%s". '
                    'Range "%s" does not end with ](+) or ](-).' % (rangeStr, subRange)
                )

            if not subRange.startswith("["):
                raise ValueError(
                    'Could not parse GenBank range string "%s". '
                    'Range "%s" does not start with "[".' % (rangeStr, subRange)
                )
            try:
                start, stop = subRange[1:-4].split(":")
            except ValueError as e:
                raise ValueError(
                    'Could not parse GenBank range string "%s". '
                    'Original parsing ValueError was "%s".' % (rangeStr, e)
                )
            else:
                if start.startswith("<"):
                    start = start[1:]
                if stop.startswith(">"):
                    stop = stop[1:]
                start, stop = map(int, (start, stop))
                if start > stop:
                    raise ValueError(
                        'Could not parse GenBank range string "%s". '
                        "Offset values (%d, %d) cannot decrease."
                        % (rangeStr, start, stop)
                    )

                ranges.append((start, stop, forward))

        self.ranges = self._mergeContiguousRanges(ranges)
        self._nRanges = len(self.ranges)

    def _mergeContiguousRanges(self, ranges):
        """
        Merge ranges that are contiguous (follow each other immediately on the
        genome).

        @param ranges: An iterable of (start, stop, forward) tuples.
        @return: A C{tuple} of (start, stop, forward) tuples with contiguous
            ranges found in C{ranges} merged.
        """
        result = []
        lastStart = lastStop = lastForward = None

        for index, (start, stop, forward) in enumerate(ranges):
            if lastStart is None:
                lastStart, lastStop, lastForward = start, stop, forward
            else:
                if start == lastStop and forward == lastForward:
                    # This range continues the previous one.
                    warn(
                        "Contiguous GenBank ranges detected: [%d:%d] "
                        "followed by [%d:%d]." % (lastStart, lastStop, start, stop)
                    )
                    lastStop = stop
                else:
                    # Emit the range that just got terminated.
                    result.append((lastStart, lastStop, lastForward))
                    # And remember this one.
                    lastStart, lastStop, lastForward = start, stop, forward

        # Emit the final range.
        result.append((lastStart, lastStop, lastForward))

        return tuple(result)

    def __str__(self):
        return "<%s: %s>" % (self.__class__.__name__, ", ".join(map(str, self.ranges)))

    def circular(self, genomeLength):
        """
        Determine whether the offset ranges of a protein in a genome span the
        end of the genome (indicating that the genome may be circular).

        @param genomeLength: The C{int} length of the genome.
        @return: A C{bool}, C{True} if the ranges overlap the genome end.
        """
        if self._nRanges == 1:
            # If there is only one range, we simply return False even though it
            # could be that the single range spans the whole genome. That is
            # not the same as having two ranges, one of which ends at the end
            # of the genome and the next starting at the start of the genome.
            #
            # I.e., a range tuples like this ((0, 100, False),) on a genome of
            # length 100 will return False, whereas ((0, 75, False), (75, 100,
            # False)) on a genome of length 100 will return True.
            #
            # The decision about whether you consider the degenerate case of
            # ((0, 100, False),) to indicate circularity is a bit arbitrary.
            # Here I've decided that it should not because I intend to use
            # this function to decide whether to output multiple SAM lines
            # for the match of a read against a reference. In this
            # degenerate case just a single SAM line suffices, whereas in a
            # non-degenerate case (such as ((0, 75, False), (75, 100, False))
            # on a genome of length 100), a chimeric SAM alignment must be
            # used, which requires quite different processing than a normal
            # linear match. YMMV, and if you don't like it you can easily
            # add an argument to this function that changes this behaviour
            # (leaving the default to do what the code currently does).
            return False

        for index, (start, stop, _) in enumerate(self.ranges):
            if stop == genomeLength:
                return self.ranges[(index + 1) % self._nRanges][0] == 0

        return False

    def startInGenome(self, match):
        """
        Calculate the position in the nucleotide genome where a DIAMOND match
        begins.

        @param match: A C{dict} with information about the DIAMOND match, as
            returned by C{self._preprocessMatch}. Must contain a 'sstart' key
            giving a 1-based C{int} offset of the start of the match in the
            protein.
        @return: The C{int} offset of the start of the DIAMOND match in the
            genome.
        """

        # Calculate the start of the match in the genome, given its start in
        # the protein.
        offsetInGenome = remaining = (match["sstart"] - 1) * 3

        for start, stop, _ in self.ranges:
            rangeWidth = stop - start
            if remaining < rangeWidth:
                # The match starts in this range.
                return start + remaining
            else:
                remaining -= rangeWidth
        else:
            raise ValueError(
                "Starting nucleotide offset %d not found in protein "
                "nucleotide ranges %s."
                % (
                    offsetInGenome,
                    ", ".join(("(%d, %d)" % (i, j)) for i, j, _ in self.ranges),
                )
            )

    def orientations(self):
        """
        Produce the set of all orientations for our ranges.

        @return: A C{set} of C{True} and C{False} range orientations.
        """
        return set(r[2] for r in self.ranges)

    def distinctRangeCount(self, genomeLength):
        """
        Determine the number of distinct ranges, given the genome length.

        @param genomeLength: The C{int} length of the genome.
        @return: An C{int} which is the number of ranges that are not
            contiguous with one another.
        """
        return len(self.ranges) - int(self.circular(genomeLength))


def getSourceInfo(genome, keys=("host", "note", "organism", "mol_type"), logfp=None):
    """
    Extract summary information from a genome source feature.

    @param genome: A GenBank genome record, as parsed by SeqIO.parse
    @param keys: An iterable of C{str} keys to extract from the Genbank
        source features.
    @param logfp: If not C{None}, a file pointer to write verbose
        progress output to.
    @return: A C{dict} with keys for the various pieces of information
        (if any) found in the source feature (see the return value below
        for detail). Or C{None} if no source feature is found or a source
        feature does not have length 1.
    """
    result = {}

    for feature in genome.features:
        if feature.type == "source":
            for key in keys:
                try:
                    values = feature.qualifiers[key]
                except KeyError:
                    value = None
                    if key != "note" and logfp:
                        print(
                            "Genome %r (accession %s) source info has "
                            "no %r feature." % (genome.description, genome.id, key),
                            file=logfp,
                        )
                else:
                    if len(values) == 1:
                        value = values[0]

                        if key == "mol_type":
                            assert value[-3:] in ("DNA", "RNA")

                    elif len(values) > 1 and key == "host":
                        value = ", ".join(values)
                    else:
                        if logfp:
                            print(
                                "Genome %r (accession %s) has source "
                                "feature %r with length != 1: %r"
                                % (genome.description, genome.id, key, values),
                                file=logfp,
                            )
                        return

                result[key] = value

            # Break so no other features are examined; we only want source.
            break
    else:
        if logfp:
            print(
                "Genome %r (accession %s) had no source feature! "
                "Skipping." % (genome.description, genome.id),
                file=logfp,
            )
        return

    return result


def getCDSInfo(
    genome, feature, keys=("gene", "note", "product", "protein_id", "translation")
):
    """
    Extract summary information from a genome CDS feature.

    @param genome: A GenBank genome record, as parsed by SeqIO.parse
    @param feature: A feature from a genome, as produced by BioPython's
        GenBank parser.
    @param keys: An iterable of C{str} keys to extract from the Genbank
        feature.
    @return: A C{dict} with keys for the various pieces of information
        found in the feature (see the return value below for detail).
        Or C{None} if the feature is not of interest or otherwise invalid.
    """
    qualifiers = feature.qualifiers

    # Check in advance that all feature qualifiers we're interested in
    # have the right lengths, if they're present.
    for key in "gene", "note", "product", "protein_id", "translation":
        if key in qualifiers:
            assert len(qualifiers[key]) == 1, (
                "GenBank qualifier key %s is not length one %r" % (key, qualifiers[key])
            )

    # A protein id is mandatory.
    if "protein_id" in qualifiers:
        proteinId = qualifiers["protein_id"][0]
    else:
        if "translation" in qualifiers:
            warn(
                "Genome %r (accession %s) has CDS feature with no "
                "protein_id feature but has a translation! "
                "Skipping.\nFeature: %s" % (genome.description, genome.id, feature)
            )
        return

    # A translated (i.e., amino acid) sequence is mandatory.
    if "translation" in qualifiers:
        translation = qualifiers["translation"][0]
    else:
        warn(
            "Genome %r (accession %s) has CDS feature with protein "
            "%r with no translated sequence. Skipping."
            % (genome.description, genome.id, proteinId)
        )
        return

    featureLocation = str(feature.location)

    # Make sure the feature's location string can be parsed.
    try:
        ranges = GenomeRanges(featureLocation)
    except ValueError as e:
        warn(
            "Genome %r  (accession %s) contains unparseable CDS "
            "location for protein %r. Skipping. Error: %s"
            % (genome.description, genome.id, proteinId, e)
        )
        return
    else:
        # Does the protein span the end of the genome? This indicates a
        # circular genome.
        circular = int(ranges.circular(len(genome.seq)))

    if feature.location.start >= feature.location.end:
        warn(
            "Genome %r (accession %s) contains feature with start "
            "(%d) >= stop (%d). Skipping.\nFeature: %s"
            % (
                genome.description,
                genome.id,
                feature.location.start,
                feature.location.end,
                feature,
            )
        )
        return

    strand = feature.strand
    if strand is None:
        # The strands of the protein in the genome are not all the same
        # (see Bio.SeqFeature.CompoundLocation._get_strand).  The
        # protein is formed by the combination of reading one strand in
        # one direction and the other in the other direction.
        #
        # This occurs just once in all 1.17M proteins found in all 700K
        # RVDB (C-RVDBv15.1) genomes, for protein YP_656697.1 on the
        # Ranid herpesvirus 1 strain McKinnell genome (NC_008211.1).
        #
        # This situation makes turning DIAMOND protein output into
        # SAM very complicated because a match on such a protein
        # cannot be stored as a SAM linear alignment. It instead
        # requires a multi-line 'supplementary' alignment. The code
        # and tests for that are more complex than I want to deal
        # with at the moment, just for the sake of one protein in a
        # frog herpesvirus.
        warn(
            "Genome %s (accession %s) has protein %r with mixed "
            "orientation!" % (genome.description, genome.id, proteinId)
        )
        return
    elif strand == 0:
        # This never occurs for proteins corresponding to genomes in
        # the RVDB database C-RVDBv15.1.
        warn(
            "Genome %r (accession %s) has protein %r with feature "
            "with strand of zero!" % (genome.description, genome.id, proteinId)
        )
        return
    else:
        assert strand in (1, -1)
        forward = strand == 1
        # Make sure the strand agrees with the orientations in the
        # string BioPython makes out of the locations.
        assert ranges.orientations() == {forward}

    return {
        "circular": circular,
        "featureLocation": featureLocation,
        "forward": forward,
        "gene": qualifiers.get("gene", [""])[0],
        "note": qualifiers.get("note", [""])[0],
        "product": qualifiers.get("product", ["UNKNOWN"])[0],
        "proteinId": proteinId,
        "ranges": ranges,
        "strand": strand,
        "translation": translation,
    }



from typing import Iterable

from prseq import FastaReader

from dark import File
from dark.reads import Reads, SSAARead
from dark.utils import asHandle


class SSFastaReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to reads in FASTA format
    along with their PDB secondary structure.

    The pdb file is in FASTA format. For each pair of sequence / structure,
    it gives the amino acid sequence in the first record and the predicted
    secondary structure in the next. This is the file format used to hold
    results from the DSSP program. Secondary structures found by DSSP can be
    downloaded from http://www.rcsb.org/pdb/files/ss.txt on 11/11/2015

    @param _files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in PDB FASTA format (see above).
    @param readClass: The class of read that should be yielded by iter. This
        must accept 3 C{str} arguments: an id, the sequence, the structure.
    @param upperCase: If C{True}, both read and structure sequences will be
        converted to upper case.
    """

    def __init__(
        self, _files: Iterable[File], readClass=SSAARead, upperCase: bool = False
    ):
        if isinstance(_files, tuple):
            self._files = _files
        elif isinstance(_files, list):
            self._files = tuple(_files)
        else:
            self._files = (_files,)

        self._readClass = readClass
        self._upperCase = upperCase
        super().__init__()

    def iter(self):
        """
        Iterate over the sequences in self.file_, yielding each as an
        instance of the desired read class.

        @raise ValueError: If the input file has an odd number of records or
            if any sequence has a different length than its predicted
            secondary structure.
        """
        upperCase = self._upperCase
        for _file in self._files:
            assert isinstance(_file, File)
            with asHandle(_file) as fp:
                records = FastaReader(fp)
                while True:
                    try:
                        record = next(records)
                    except StopIteration:
                        break
                    try:
                        structureRecord = next(records)
                    except StopIteration:
                        raise ValueError(
                            "Structure file %r has an odd number of records." % _file
                        )

                    if len(structureRecord.sequence) != len(record.sequence):
                        raise ValueError(
                            "Sequence %r length (%d) is not equal to "
                            "structure %r length (%d) in input file %r."
                            % (
                                record.id,
                                len(record.sequence),
                                structureRecord.id,
                                len(structureRecord.sequence),
                                _file,
                            )
                        )

                    if upperCase:
                        read = self._readClass(
                            record.id,
                            str(record.sequence.upper()),
                            str(structureRecord.sequence.upper()),
                        )
                    else:
                        read = self._readClass(
                            record.id,
                            str(record.sequence),
                            str(structureRecord.sequence),
                        )

                    yield read



#!/usr/bin/env python

import argparse
import sys

from ete4 import Tree

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Print a Newick phylogenetic tree in ASCII.",
)

parser.add_argument(
    "--treeFile",
    metavar="FILENAME",
    default=sys.stdin,
    type=open,
    help="The Newick tree file.",
)

parser.add_argument(
    "--outgroupRegex",
    metavar="TAXON-REGEX",
    help=(
        "A regular expression (for taxa names). The lowest common parent of "
        "any matching taxa will be used as an outgroup."
    ),
)

parser.add_argument(
    "--parser",
    "--format",
    metavar="N",
    type=int,
    help="The ete4 Newick parser to use.",
)

parser.add_argument(
    "--verbose",
    action="store_true",
    help="Print information about the outgroup (if any) taxa to standard error.",
)

args = parser.parse_args()

tree = Tree(args.treeFile.read(), parser=args.parser)

if args.outgroupRegex:
    from re import compile

    regex = compile(args.outgroupRegex)
    taxa = [leaf.name for leaf in tree.iter_leaves() if regex.match(leaf.name)]

    if taxa:
        ca = tree.get_common_ancestor(taxa)
        if args.verbose:
            print("Taxa for outgroup:", taxa, file=sys.stderr)
            print("Common ancestor:", ca.name, file=sys.stderr)
            print("Common ancestor is tree:", tree == ca, file=sys.stderr)

        if len(taxa) == 1:
            tree.set_outgroup(tree & taxa[0])
        else:
            if ca == tree:
                tree.set_outgroup(tree.get_midpoint_outgroup())
            else:
                tree.set_outgroup(tree.get_common_ancestor(taxa))

print(tree.to_str())



#!/usr/bin/env python

import sys
from argparse import ArgumentParser

from dark.aa import find
from dark.aaVars import (
    ALL_PROPERTIES,
    CODONS,
    NAMES_TO_ABBREV1,
    PROPERTY_NAMES,
    STOP_CODONS,
)

parser = ArgumentParser()
parser.add_argument("targets", nargs="*", help="The items to look up.")
parser.add_argument(
    "--details", action="store_true", help="Display numeric amino acid property values."
)
args = parser.parse_args()

targets = args.targets or sorted(NAMES_TO_ABBREV1)

error = False

for target in targets:
    aas = list(find(target))
    if aas:
        for aa in aas:
            print(aa.name)
            print("  3-letter abbreviation: %s" % aa.abbrev3)
            print("  1-letter abbreviation: %s" % aa.abbrev1)
            print("  Codons: %s" % ", ".join(sorted(aa.codons)))

            properties = []
            print("  Properties:", end=" ")
            for prop in ALL_PROPERTIES:
                if aa.properties & prop:
                    properties.append(PROPERTY_NAMES[prop])
            print(", ".join(properties))

            if args.details:
                print("  Property details:")
                for propertyDetail, value in aa.propertyDetails.items():
                    print("    %s: %s" % (propertyDetail, value))
    else:
        if (
            target.upper() in STOP_CODONS
            or target.upper().replace("U", "T") in STOP_CODONS
        ):
            print("Stop codon.")
        else:
            error = True
            print("Unknown amino acid or codon: %s" % target, file=sys.stderr)

if error:
    print("Valid AA arguments are: %s." % list(CODONS), file=sys.stderr)
    sys.exit(1)



#!/usr/bin/env python

import argparse

from dark.sam import samReferencesToStr

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Print SAM/BAM file reference names and lengths.",
)

parser.add_argument(
    "samfile", metavar="FILENAME", help="The name of a SAM/BAM alignment file."
)

print(samReferencesToStr(parser.parse_args().samfile))



#!/usr/bin/env python

import argparse
import sys
from math import log10
from os.path import join

from dark.aa import compareAaReads, matchToString
from dark.fasta import FastaReads
from dark.process import Executor
from dark.reads import Reads, addFASTACommandLineOptions, parseFASTACommandLineOptions
from dark.utils import parseRangeExpression


def needle(reads):
    """
    Run a Needleman-Wunsch alignment and return the two sequences.

    @param reads: An iterable of two reads.
    @return: A C{Reads} instance with the two aligned sequences.
    """
    from shutil import rmtree
    from tempfile import mkdtemp

    dir = mkdtemp()

    file1 = join(dir, "file1.fasta")
    with open(file1, "w") as fp:
        print(reads[0].toString("fasta"), end="", file=fp)

    file2 = join(dir, "file2.fasta")
    with open(file2, "w") as fp:
        print(reads[1].toString("fasta"), end="", file=fp)

    out = join(dir, "result.fasta")

    Executor().execute(
        "needle -asequence '%s' -bsequence '%s' -auto "
        "-outfile '%s' -aformat fasta" % (file1, file2, out)
    )

    # Use 'list' in the following to force reading the FASTA from disk.
    result = Reads(list(FastaReads(out)))
    rmtree(dir)

    return result


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=("Compare two amino acid sequences."),
)

parser.add_argument(
    "--index1",
    type=int,
    default=1,
    help="The (1-based) index in the input of the first sequence.",
)

parser.add_argument(
    "--index2",
    type=int,
    default=2,
    help="The (1-based) index in the input of the second sequence.",
)

parser.add_argument(
    "--align",
    default=False,
    action="store_true",
    help="If given, use needle to do a global alignment of the two sequences.",
)

parser.add_argument(
    "--alignmentFile", help="The file to save the alignment to (implies --align)."
)

parser.add_argument(
    "--sites",
    help=(
        "Specify (1-based) sequence sites to keep. All other sites will "
        "be ignored. The sites must be given in the form e.g., "
        "24,100-200,260."
    ),
)

parser.add_argument(
    "--showDiffs",
    default=False,
    action="store_true",
    help="Print (1-based) sites where the sequence amino acids differ.",
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()

keepSequences = set([args.index1 - 1, args.index2 - 1])

reads = list(parseFASTACommandLineOptions(args).filter(keepSequences=keepSequences))

if len(reads) == 1:
    if len(keepSequences) == 1:
        # This is ok, they want to compare a sequence with itself.
        reads = Reads([reads[0], reads[0]])
    else:
        print("Could not find both requested sequence indices. Exiting.")
        sys.exit(1)
elif len(reads) != 2:
    print("Could not find both requested sequence indices. Exiting.")
    sys.exit(1)

if args.alignmentFile:
    args.align = True

if args.align:
    len1, len2 = map(len, reads)
    if len1 == len2:
        print("Pre-alignment, sequence lengths were identical: %s" % len1)
    else:
        print(
            "Pre-alignment, sequence lengths: %d, %d (difference %d)"
            % (len1, len2, abs(len1 - len2))
        )

    # Align.
    reads = needle(reads)

    if args.alignmentFile:
        assert reads.save(args.alignmentFile) == 2

offsets = (
    parseRangeExpression(args.sites, convertToZeroBased=True) if args.sites else None
)

read1, read2 = reads
len1, len2 = map(len, reads)
identicalLengths = len1 == len2

# Sanity check.
if args.align:
    assert identicalLengths

match = compareAaReads(read1, read2, offsets=offsets)

x = "Post-alignment, sequence" if args.align else "Sequence"
if identicalLengths:
    print("%s lengths are identical: %s" % (x, len1))
else:
    print("%s lengths: %d, %d (difference %d)" % (x, len1, len2, abs(len1 - len2)))

print(matchToString(match, read1, read2, offsets=offsets))

if args.showDiffs:
    # Print all sites where the sequences differ.
    width = int(log10(max(len1, len2))) + 1
    headerPrinted = False
    for site, (a, b) in enumerate(zip(read1.sequence, read2.sequence), start=1):
        if a != b:
            if not headerPrinted:
                print("Differences (site, %s, %s):" % (read1.id, read2.id))
                headerPrinted = True
            print("  %*d %s %s" % (width, site, a, b))

    if not headerPrinted:
        print("No sequence differences found.")



import multiprocessing
import os
from itertools import product
from os.path import join
from tempfile import mkdtemp
from typing import TextIO, final

import requests

from dark.process import Executor
from dark.sam import samfile as openSamfile


@final
class Bowtie2:
    """
    Run Bowtie2.
    """

    def __init__(
        self,
        executor: Executor | None = None,
        threads: int | None = None,
        verboseFp: TextIO | None = None,
        dryRun: bool = False,
        reference: str | None = None,
        tempdir: str | None = None,
        tmpChmod: str | None = None,
    ) -> None:
        self._executor = executor or Executor(dryRun)
        if dryRun:
            self.tempdir = tempdir or "/tmp/xxx"
        else:
            self.tempdir = mkdtemp(prefix="bt2-", dir=tempdir)
            if tmpChmod:
                self._executor.execute(f"chmod {tmpChmod} {self.tempdir}")
        self._samFile = join(self.tempdir, "result.sam")
        self._bamFile = join(self.tempdir, "result.bam")
        self._indexFile = join(self.tempdir, "index")
        self._verboseFp = verboseFp
        self._indexCalled = self._samExists = self._bamExists = False
        self.nThreads = threads or multiprocessing.cpu_count()
        self._reference = reference

    def buildIndex(self, index: str) -> None:
        """
        Find or make a Bowtie2 index.
        """
        if os.path.exists(index):
            # Check if this is a pre-existing bowtie2 index. Look for and remove a
            # bowtie2 suffix, if any (otherwise bowtie2 will complain). We do things
            # this way to allow a user to use TAB completion on the command line to give
            # them the full path to any bowtie index file.
            #
            # Note that Bowtie2 puts an "l" on the end of the filename of its 'large'
            # indices.
            for suffix1, suffix2 in product(
                ("1", "2", "3", "4", "rev.1", "rev.2"),
                ("", "l"),
            ):
                suffix = f".{suffix1}.bt2{suffix2}"
                if index.endswith(suffix):
                    self._indexFile = index[: -len(suffix)]
                    self._report(
                        f"Using pre-existing Bowtie2 index {self._indexFile!r}."
                    )
                    break
            else:
                # Assume a FASTA file and make an index.
                self._indexFile = self._makeIndexFromFastaFile(index)
        else:
            # Not a filename. So either the start of the path to a bowtie2 index or else
            # an accession number. Bowtie2 puts an "l" on the end of the filename of its
            # 'large' indices, so we check for that too.
            if os.path.exists(index + ".1.bt2") or os.path.exists(index + ".1.bt2l"):
                self._report(f"Using pre-existing Bowtie2 index {index!r}.")
                self._indexFile = index
            else:
                # Assume an accession number.
                self._report(
                    f"Treating index argument {index!r} as an accession number."
                )
                self._indexFile = self._makeIndexFromAccession(index)

        self._indexCalled = True

    def align(
        self,
        bowtie2Args: str = "--no-unal",
        fastq1: str | None = None,
        fastq2: str | None = None,
        threads: int | None = None,
        discardSAM: bool = False,
        readGroup: str = "orig",
        sampleName: str = "orig",
    ) -> None:
        """
        Run Bowtie2.
        """
        if not self._indexCalled:
            raise ValueError("buildIndex() has not yet been called.")

        self._report("Aligning with Bowtie2.")

        nThreads = threads or self.nThreads
        samFile = "/dev/null" if discardSAM else self._samFile

        if fastq1 and fastq2:
            self._executor.execute(
                "bowtie2 %s --threads %d --rg-id '%s' --rg 'SM:%s' -x '%s' "
                "-1 '%s' -2 '%s' > '%s'"
                % (
                    bowtie2Args,
                    nThreads,
                    readGroup,
                    sampleName,
                    self._indexFile,
                    fastq1,
                    fastq2,
                    samFile,
                )
            )
        elif fastq1:
            self._executor.execute(
                "bowtie2 %s --threads %d --rg-id '%s' --rg 'SM:%s' -x '%s' "
                "-U '%s' > '%s'"
                % (
                    bowtie2Args,
                    nThreads,
                    readGroup,
                    sampleName,
                    self._indexFile,
                    fastq1,
                    samFile,
                )
            )
        else:
            raise ValueError("At least fastq1 must be passed.")

        self._samExists = True

    def _report(self, mesg: str) -> None:
        """
        Print a progress message.
        """
        if self._verboseFp:
            print(mesg, file=self._verboseFp)

    def makeBAM(self, samtoolsViewArgs: str = "") -> None:
        """
        Convert the SAM to BAM.
        """
        # The following will raise if there is no SAM file.
        self._SAMorBAM()
        self._report("Converting SAM to BAM.")
        self._executor.execute(
            "samtools view -b %s '%s' > '%s'"
            % (samtoolsViewArgs, self._samFile, self._bamFile)
        )
        self._bamExists = True

    def _SAMorBAM(self) -> str:
        """
        Do we have a SAM or BAM file available?
        """
        if self._bamExists:
            return "BAM"
        elif self._samExists:
            return "SAM"
        else:
            raise ValueError("bowtie2() has not yet been called.")

    def outputFile(self) -> str:
        """
        The name of the output file.
        """
        if self._bamExists:
            return self._bamFile
        elif self._samExists:
            return self._samFile
        else:
            raise ValueError("bowtie2() has not yet been called.")

    def indexBAM(self) -> None:
        """
        Index the BAM file.
        """
        which = self._SAMorBAM()

        if which != "BAM":
            raise ValueError("makeBAM() has not yet been called.")

        self._report("Indexing BAM.")
        self._executor.execute("samtools index '%s'" % self._bamFile)

    def isEmpty(self) -> bool:
        """
        Are any reads present in the SAM file?
        """
        # Reading the SAM file manually doesn't require BAM that is sorted and indexed
        # (which is what we would need to do if we wanted to use the mapped and unmapped
        # attributes of an AlignmentFile instance).
        with open(self._samFile) as fp:
            for line in fp:
                if not line.startswith("@"):
                    return False
        return True

    def mappedAndUnmappedCounts(self) -> tuple[int, int]:
        """
        Return the number of mapped and unmapped reads.

        @return: A 2-C{tuple} of C{int}s, giving the number of mapped and unmapped
            reads.
        """
        which = self._SAMorBAM()

        if which != "BAM":
            raise ValueError("makeBAM() has not yet been called.")

        with openSamfile(self._bamFile) as samfile:
            # This will fail if the BAM file has not been sorted and indexed.
            return samfile.mapped, samfile.unmapped

    def sort(self, byName: bool = False) -> None:
        """
        Sort the BAM or SAM.
        """
        which = self._SAMorBAM()
        self._report("Sorting %s (by %s)." % (which, "name" if byName else "coord"))
        inFile = self._bamFile if which == "BAM" else self._samFile
        sortedFile = join(self.tempdir, "result-sorted." + which.lower())
        self._executor.execute(
            "samtools sort --output-fmt %s %s'%s' > '%s'"
            % (which, "-n " if byName else "", inFile, sortedFile)
        )
        self._executor.execute("mv '%s' '%s'" % (sortedFile, inFile))

    def removePrimers(self, bedFile: str) -> None:
        """
        Removes primers specified in the bed file
        """
        which = self._SAMorBAM()

        if which != "BAM":
            raise ValueError("makeBAM() has not yet been called.")

        self._report("removing primers specified in %s" % bedFile)
        tempTrimmedBamPrefix = "%s.trimmed" % self._bamFile
        self._executor.execute(
            "ivar trim -b '%s' -p '%s' -i '%s' -q 20 -m 30 -s 4 -e"
            % (bedFile, tempTrimmedBamPrefix, self._bamFile)
        )
        self._executor.execute(
            "mv '%s'.bam '%s'" % (tempTrimmedBamPrefix, self._bamFile)
        )

    def markDuplicatesPicard(self, picardFile: str) -> None:
        """
        Use Picard to mark duplicates.
        """
        which = self._SAMorBAM()
        self._report("Marking duplicates with Picard.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "picard-duplicates." + which.lower())
        tempErrFile = join(self.tempdir, "picard.errs")

        self._executor.execute(
            "java -Xmn2g -Xms2g -Xmx2g -jar %s "
            "MarkDuplicates I='%s' O='%s' M=/dev/null >'%s' 2>&1"
            % (picardFile, inFile, tempFile, tempErrFile)
        )

        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def markDuplicatesGATK(self, threads: int | None = None) -> None:
        """
        Use GATK to mark duplicates.
        """
        nThreads = threads or self.nThreads
        which = self._SAMorBAM()
        self._report("Marking duplicates with GATK.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "gatk-duplicates." + which.lower())

        self._executor.execute(
            "gatk MarkDuplicatesSpark "
            "-I '%s' -O '%s' --conf spark.executor.cores=%d"
            % (inFile, tempFile, nThreads)
        )

        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def callHaplotypesGATK(
        self,
        picardJar: str,
        vcfFile: str | None = None,
        referenceFasta: str | None = None,
    ) -> None:
        """
        Use GATK to call haplotypes.
        """
        which = self._SAMorBAM()
        self._report("Calling haplotypes with GATK.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        vcfFile = vcfFile or join(self.tempdir, "output.vcf.gz")

        if referenceFasta is None:
            if self._reference:
                self._report("Using %s as a reference." % self._reference)
                referenceFasta = self._reference
            else:
                raise ValueError(
                    "No reference was passed, given to the "
                    "Bowtie2 __init__, or used in buildIndex."
                )

        indexFile = referenceFasta + ".fai"
        if os.path.exists(indexFile):
            removeIndex = False
        else:
            removeIndex = True
            self._executor.execute("samtools faidx '%s'" % referenceFasta)

        if referenceFasta.lower().endswith(".fasta"):
            dictFile = referenceFasta[: -len(".fasta")] + ".dict"
        else:
            dictFile = referenceFasta + ".dict"

        if os.path.exists(dictFile):
            removeDict = False
        else:
            removeDict = True
            self._executor.execute(
                "java -jar '%s' CreateSequenceDictionary R='%s' O='%s'"
                % (picardJar, referenceFasta, dictFile)
            )

        self._executor.execute(
            "gatk --java-options -Xmx4g HaplotypeCaller "
            "--reference '%s' "
            "--input '%s' "
            "--output '%s' "
            "--sample-ploidy 1 "
            "--dont-use-soft-clipped-bases true "
            "-ERC GVCF" % (referenceFasta, inFile, vcfFile)
        )

        if removeIndex:
            self._executor.execute("rm '%s'" % indexFile)

        if removeDict:
            self._executor.execute("rm '%s'" % dictFile)

    def callHaplotypesBcftools(
        self, vcfFile: str | None = None, referenceFasta: str | None = None
    ) -> None:
        """
        Use bcftools call to call haplotypes.
        """
        which = self._SAMorBAM()
        self._report("Calling haplotypes with bcftools call.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        vcfFile = vcfFile or join(self.tempdir, "output.vcf.gz")

        if referenceFasta is None:
            if self._reference:
                self._report("Using %s as a reference." % self._reference)
                referenceFasta = self._reference
            else:
                raise ValueError(
                    "No reference was passed, given to the "
                    "Bowtie2 __init__, or used in buildIndex."
                )

        self._executor.execute(
            'bcftools mpileup --max-depth 5000 -Ou -f "%s" "%s" | '
            'bcftools call --ploidy 1 -mv -Oz -o "%s"'
            % (referenceFasta, inFile, vcfFile)
        )

        self._executor.execute(f"bcftools index --force {vcfFile!r}")

    def removeDuplicates(self) -> None:
        """
        Use samtools to remove marked duplicates.
        """
        which = self._SAMorBAM()
        self._report("Removing marked duplicates.")

        inFile = self._bamFile if which == "BAM" else self._samFile
        tempFile = join(self.tempdir, "non-duplicates." + which.lower())

        # See the comment on the testRemoveDuplicates test in
        # ../test/test_bowtie2.py regarding there being two spaces in the
        # samtools command below if we are not producing BAM.  If you
        # change spacing here you will also need to change that test.
        bArg = "-b" if which == "BAM" else ""
        self._executor.execute(
            "samtools view %s -F 1024 '%s' > '%s'" % (bArg, inFile, tempFile)
        )
        self._executor.execute("mv '%s' '%s'" % (tempFile, inFile))

    def _makeIndexFromFastaFile(self, fastaFilename: str) -> str:
        index = join(self.tempdir, "index")

        self._report("Building Bowtie2 index from %s." % fastaFilename)
        self._executor.execute(
            "bowtie2-build --quiet '%s' '%s'" % (fastaFilename, index)
        )

        if self._reference is None:
            self._report("Will use this FASTA file as the reference (if needed).")
            self._reference = fastaFilename

        return index

    def _makeIndexFromAccession(self, accessionId: str) -> str:
        fastaFilename = join(self.tempdir, accessionId + ".fasta")
        index = join(self.tempdir, "index")

        self._report(f"Downloading FASTA for accession {accessionId!r} from NCBI.")

        URL = (
            "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
            f"db=nucleotide&id={accessionId}&rettype=fasta&retmode=text"
        )

        if not self._executor.dryRun:
            with open(fastaFilename, "w") as fp:
                print(requests.get(URL).text.rstrip("\n"), file=fp)

        self._report("Building bowtie2 index for {accessionId!r}.")
        self._executor.execute(f"bowtie2-build --quiet {fastaFilename!r} {index!r}")

        return index

    def close(self) -> None:
        """
        Clean up.
        """
        self._executor.execute("rm -r '%s'" % self.tempdir)



#!/usr/bin/env python

import argparse

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


def main():
    parser = argparse.ArgumentParser(
        description="Given FASTA on stdin, write the sequence ids to stdout.",
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    for read in reads:
        print(read.id)


if __name__ == "__main__":
    main()



import sys
from collections.abc import Iterable

from prseq import FastqReader

from dark import File
from dark.reads import DNARead, Reads


class FastqReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to FASTQ reads.

    @param files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in FASTQ format.
    @param readClass: The class of read that should be yielded by iter.
    """

    def __init__(self, files: File | Iterable[File] | None = None, readClass=DNARead):
        if files is None:
            self.files = (sys.stdin.buffer,)
        elif isinstance(files, (list, tuple)):
            self.files = tuple(files)
        else:
             self.files = (files,)
        self.readClass = readClass
        super().__init__()

    def iter(self):
        """
        Iterate over the sequences in the files in self.files, yielding each
        as an instance of the desired read class.
        """
        for _file in self.files:
            for record in FastqReader(_file):
                yield self.readClass(record.id, record.sequence, record.quality)



from unittest import TestCase, skipUnless

from dark.consensus import ConsensusError, consensusFromBAM
from dark.reads import DNARead
from dark.sam import (
    UnequalReferenceLengthError,
    UnknownReference,
    UnspecifiedReference,
    samtoolsInstalled,
)

from .bam import REF_ID, makeBAM


@skipUnless(samtoolsInstalled(), "samtools is not installed")
class TestReferenceErrors(TestCase):
    """
    Test expected errors.
    """

    def testUnknownStrategy(self):
        """
        If an unknown strategy is passed, ValueError must be raised.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            strategy = "xxx"
            error = rf"^Unknown consensus strategy {strategy!r}\.$"
            self.assertRaisesRegex(
                ConsensusError,
                error,
                consensusFromBAM,
                bamFilename,
                strategy=strategy,
                quiet=True,
                referenceFasta=fastaFilename,
            )

    def testUnknownReferenceId(self):
        """
        If a reference id that is not in the BAM file is passed,
        UnknownReference must be raised.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            badId = "bad-id"
            error = (
                rf"^BAM file {str(bamFilename)!r} does not mention a "
                rf"reference with id {badId!r}\. Known references are: "
                rf"ref-id\.$"
            )
            self.assertRaisesRegex(
                UnknownReference,
                error,
                consensusFromBAM,
                bamFilename,
                bamId=badId,
                referenceFasta=fastaFilename,
            )

    def testUnknownReference(self):
        """
        If a reference sequence is passed and its id is not in the BAM
        file, UnknownReference must be raised.
        """
        template = ("ACGTTCCG",)

        UNKNOWN = "unknown-id"
        with makeBAM(template) as (fastaFilename, bamFilename):
            error = (
                rf"^BAM file {str(bamFilename)!r} does not mention a "
                rf"reference with id {UNKNOWN!r}\. Known "
                rf"references are: ref-id\.$"
            )
            self.assertRaisesRegex(
                UnknownReference,
                error,
                consensusFromBAM,
                bamFilename,
                bamId=UNKNOWN,
                referenceFasta=fastaFilename,
            )

    def testUnknownFastaReference(self):
        """
        If a reference sequence id is passed and it is not in the FASTA file,
        UnknownReference must be raised.
        """
        template = ("ACGTTCCG",)

        UNKNOWN = "unknown-id"
        with makeBAM(template) as (fastaFilename, bamFilename):
            error = (
                rf"^No sequence with id {UNKNOWN!r} found in "
                rf"{str(fastaFilename)!r}\.$"
            )
            self.assertRaisesRegex(
                UnknownReference,
                error,
                consensusFromBAM,
                bamFilename,
                fastaId=UNKNOWN,
                referenceFasta=fastaFilename,
            )

    def testUnspecifiedReference(self):
        """
        If a reference is not specified and the BAM file mentions more than
        one, UnspecifiedReference must be raised.
        """
        template = ("ACGTTCCG",)

        bamReferences = [DNARead("ref-1", template[0]), DNARead("ref-2", "AA")]
        fastaReferences = [DNARead("ref-3", "AAA")]
        with makeBAM(
            template, bamReferences=bamReferences, fastaReferences=fastaReferences
        ) as (fastaFilename, bamFilename):
            error = (
                r"^Could not infer a BAM reference. Available references are: "
                r"ref-1, ref-2\.$"
            )
            self.assertRaisesRegex(
                UnspecifiedReference,
                error,
                consensusFromBAM,
                bamFilename,
                referenceFasta=fastaFilename,
            )

    def testReferenceOfWrongLength(self):
        """
        If a reference sequence is passed and it does not have the right
        length, UnequalReferenceLengthError must be raised.
        """
        template = ("ACGTTCCG",)

        fastaReferences = [DNARead(REF_ID, "AA")]
        with makeBAM(template, fastaReferences=fastaReferences) as (
            fastaFilename,
            bamFilename,
        ):
            error = (
                rf"^Reference FASTA sequence {REF_ID!r} has length 2, but the "
                rf"BAM reference {REF_ID!r} has length {len(template[0])}\.$"
            )
            self.assertRaisesRegex(
                UnequalReferenceLengthError,
                error,
                consensusFromBAM,
                bamFilename,
                referenceFasta=fastaFilename,
                quiet=True,
            )

    def testEmptyReferenceFile(self):
        """
        If a reference file is passed but is empty, UnknownReference must be
        raised.
        """
        template = ("ACGTTCCG",)

        fastaReferences = []
        with makeBAM(template, fastaReferences=fastaReferences) as (
            fastaFilename,
            bamFilename,
        ):
            error = (
                rf"^The FASTA reference file {str(fastaFilename)!r} contained "
                rf"no sequences\.$"
            )
            self.assertRaisesRegex(
                UnknownReference,
                error,
                consensusFromBAM,
                bamFilename,
                referenceFasta=fastaFilename,
            )


class _Mixin:
    """
    Common (i.e., with and without considering FASTQ quality values) tests for
    consensuses making.
    """

    def testNoReadsReference(self):
        """
        If no reads are present and resolution of no-coverage bases is the
        reference sequence, the reference should be returned.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                template[0],
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testNoReadsReferenceFromId(self):
        """
        The reference can be passed using just its id (as opposed to a full
        reference Read instance). A string of 'N's is returned because there
        are no reads and we pass noCoverage='N'.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "N" * len(template[0]),
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    bamId="ref-id",
                    noCoverage="N",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReferenceAndReferenceIdNotGiven(self):
        """
        If the reference to use is not given, the single reference in the BAM
        file should be used. A string of 'N's is returned because there
        are no reads and we pass noCoverage='N'.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "N" * len(template[0]),
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="N",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testNoReadsN(self):
        """
        If no reads are present and resolution of no-coverage bases is 'N'
        (or '?', etc) a sequence of Ns (or ?s, etc) should be returned.
        """
        template = ("ACGTTCCG",)

        with makeBAM(template) as (fastaFilename, bamFilename):
            for char in "N?":
                self.assertEqual(
                    char * len(template[0]),
                    consensusFromBAM(
                        bamFilename,
                        quiet=True,
                        referenceFasta=fastaFilename,
                        noCoverage=char,
                        ignoreQuality=self.ignoreQuality,
                    ).sequence,
                )

    def testLowReadsReference(self):
        """
        If fewer reads than needed are present and resolution of low-coverage
        sites is the reference sequence, the reference should be returned.
        """
        template = (
            "ACGTTCCG",
            "  GTT",
            "  ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                template[0],
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    minCoverage=2,
                    lowCoverage="reference",
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testLowReadsCharNoCoverageConsensus(self):
        """
        If fewer reads than needed are present and resolution of low-coverage
        sites is 'N' (or '? etc), then Ns (or ?s, etc) should be returned in
        the low coverage sites, and the reference in the sites with no
        coverage.
        """
        template = (
            "ACGTTCCG",
            "  GTT",
            "  ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            for char in "N?":
                self.assertEqual(
                    "AC" + char * 3 + "CCG",
                    consensusFromBAM(
                        bamFilename,
                        quiet=True,
                        referenceFasta=fastaFilename,
                        minCoverage=2,
                        lowCoverage=char,
                        noCoverage="reference",
                        ignoreQuality=self.ignoreQuality,
                    ).sequence,
                )

    def testLowReadsCharNoCoverageX(self):
        """
        If fewer reads than needed are present and resolution of low-coverage
        sites is 'N' (or '? etc), then Ns (or ?s, etc) should be returned in
        the low coverage sites, and (for example) 'X' in the sites with no
        coverage.
        """
        template = (
            "ACGTTCCG",
            "  GTT",
            "  ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            for char in "N?":
                self.assertEqual(
                    "XX" + char * 3 + "XXX",
                    consensusFromBAM(
                        bamFilename,
                        quiet=True,
                        referenceFasta=fastaFilename,
                        noCoverage="X",
                        lowCoverage=char,
                        minCoverage=2,
                        ignoreQuality=self.ignoreQuality,
                    ).sequence,
                )

    def testOneReadMatchingPartOfTheReference(self):
        """
        If one read is present and it matches part of the reference and
        resolution of no-coverage bases is the reference sequence, the
        reference should be returned.
        """
        template = (
            "ACGTTCCG",
            "  GTT",
            "  ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                template[0],
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneReadDifferingFromPartOfTheReference(self):
        """
        If one read is present and it differs from part of the reference and
        resolution of no-coverage bases is the reference sequence, the
        expected hybrid of the read and the reference should be returned.
        """
        template = (
            "ACGTTCCG",
            "  AAA",
            "  ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACAAACCG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoReadsDifferingFromPartOfTheReferenceSomeLowCoverage(self):
        """
        If two reads are present and they differ from part of the reference and
        resolution of no-coverage bases is the reference sequence, the
        expected hybrid of the read and the reference should be returned, with
        the part of the reads that has insufficient coverage returning the
        low-coverage symbol (here '+').
        """
        template = (
            "ACGTTCCG",
            "  AAA",
            "  ???",
            "  AAAT",
            "  ????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACAAA+CG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    minCoverage=2,
                    noCoverage="reference",
                    lowCoverage="+",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoReadsDifferingFromPartOfTheReferenceLowAndNoCoverage(self):
        """
        If two reads are present and they differ from part of the reference and
        resolution of no-coverage bases is 'N', the expected hybrid of the read
        and the reference should be returned, with the part of the reads that
        has insufficient coverage returning the low-coverage symbol (here '+').
        """
        template = (
            "ACGTTCCG",
            "  AAA",
            "  ???",
            "  AAAT",
            "  ????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "NNAAA+NN",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    minCoverage=2,
                    noCoverage="N",
                    lowCoverage="+",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testSimpleMajority(self):
        """
        If three reads result in a majority base at a site, that base should
        be in the consensus.
        """
        template = (
            "ACGT",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  C",
            "  ?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACAT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.5,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testSimpleMajorityBelowThreshold(self):
        """
        If conflicting reads at a site do not give a simple (above threshold)
        majority, the ambiguous code should be in the consensus.
        """
        template = (
            "ACGT",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  C",
            "  ?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACMT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasEarlierSites(self):
        """
        If a read has sites that come before the reference, they must
        be included in the consensus.
        """
        template = (
            " ACGT",
            "AA",
            "??",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "AACGT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasEarlierSitesNoSoftClipped(self):
        """
        If a read has sites that come before the reference, they must
        not appear in the consensus if soft-clipped bases are not included.
        """
        template = (
            " ACGT",
            "AA",
            "??",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACGT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasLaterSites(self):
        """
        If a read has sites that come after the reference, they must
        be included in the consensus.
        """
        template = (
            "ACGT",
            "   TAA",
            "   ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACGTAA",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasLaterSitesNoSoftClipped(self):
        """
        If a read has sites that come after the reference, they must
        not appear in the consensus if soft-clipped bases are not included.
        """
        template = (
            "ACGT",
            "   TAA",
            "   ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACGT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasEarlierAndLaterSites(self):
        """
        If a read has sites that come before and after the reference, the sites
        must be included in the consensus.
        """
        template = (
            "  ACGT",
            "TCACGTGA",
            "????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TCACGTGA",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadHasEarlierAndLaterSitesNoSoftClipped(self):
        """
        If a read has sites that come before and after the reference, the sites
        must not appear in the consensus if soft-clipped bases are not
        included.
        """
        template = (
            "  ACGT",
            "TCACGTGA",
            "????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACGT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadsHaveEarlierAndLaterSites(self):
        """
        If two reads have sites that come before and after the reference, the
        sites must be included in the consensus. This is the same as the
        testReadHasEarlierAndLaterSites test above, but using two reads
        (instead of one) to give the before and after sites.
        """
        template = (
            "  ACGT",
            "TCA",
            "???",
            "     TGA",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TCACGTGA",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testReadsHaveEarlierAndLaterSitesNoSoftClipped(self):
        """
        If two reads have sites that come before and after the reference, the
        sites must not appear in the consensus if soft-clipped bases are not
        included. This is the same as the
        testReadHasEarlierAndLaterSitesNoSoftClipped test above, but using two
        reads (instead of one) to give the before and after sites.
        """
        template = (
            "  ACGT",
            "TCA",
            "???",
            "     TGA",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACGT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneDeletionFromReference(self):
        """
        A deletion from the reference must be handled correctly.
        """
        template = (
            "CACGTG",
            " A-A",
            " ?-?",
            " A-A",
            " ?-?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAxATG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="x",
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneLowFrequencyDeletion(self):
        """
        A deletion from the reference that does not meet the required deletion
        frequency should not appear in the consensus.
        """
        template = (
            "CACGTG",
            " A-A",
            " ?-?",
            " AGA",
            " ???",
            " AGA",
            " ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAGATG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="x",
                    deletionThreshold=0.5,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneDeletionFromReferenceUnmarked(self):
        """
        A deletion from the reference must be handled correctly.
        """
        template = (
            "CACGTG",
            " A-A",
            " ?-?",
            " A-A",
            " ?-?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAATG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="",
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoDeletionsFromReference(self):
        """
        Two deletions from the reference must be handled correctly.
        """
        template = (
            "CACGTG",
            " A-A-G",
            " ?-?-?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAxAxG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="x",
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoDeletionsFromReferenceUnmarked(self):
        """
        Two deletions from the reference must be handled correctly.
        """
        template = (
            "CACGTG",
            " A-A-G",
            " ?-?-?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="",
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneInsertionInReference(self):
        """
        An insertion in the reference must be handled correctly.
        """
        template = (
            "CA-CGTG",
            " AAC",
            " ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAACGTG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOneInsertionInReferenceLowFrequency(self):
        """
        An insertion in the reference must not be included if it does not
        meet the insertionCountThreshold.
        """
        template = (
            "CA-CGTG",
            " ATC",
            " ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CACGTG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    insertionCountThreshold=2,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertionsInReference(self):
        """
        Two insertions in the reference must be handled correctly.
        """
        template = (
            "CA-CGT-G",
            " AAC",
            " ???",
            "     TAG",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAACGTAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertionsAndOneUnmarkedDeletionInReference(self):
        """
        Two insertions and one deletion in the reference must be handled
        correctly.
        """
        template = (
            "CA-CGT-G",
            " AA-G",
            " ??-?",
            "     TAG",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAAGTAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="",
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertionsAndOneDeletionInReference(self):
        """
        Two insertions and one deletion in the reference must be handled
        correctly, with the deletion marked with 'x' as requested.
        """
        template = (
            "CA-CGT-G",
            " AA-G",
            " ??-?",
            "     TAG",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAAxGTAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="x",
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertionsAndTwoDeletionsInReference(self):
        """
        Two insertions and two deletions in the reference must be handled
        correctly and the deletions must appear in the consensus as requested.
        """
        template = (
            "CA-CGT-G",
            " AA--",
            " ??--",
            "     TAG",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAAxxTAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="x",
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertionsAndTwoUnmarkedDeletionsInReference(self):
        """
        Two insertions and two deletions in the reference must be handled
        correctly and the deletions must not appear in the consensus if the
        deletion symbol is the empty string.
        """
        template = (
            "CA-CGT-G",
            " AA--",
            " ??--",
            "     TAG",
            "     ???",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "CAATAG",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.7,
                    deletionSymbol="",
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testGeneiousExamplesNoTie(self):
        """
        Test the no-tied counts example from
        https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
        """
        template = (
            "ACGT",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  G",
            "  ?",
            "  G",
            "  ?",
            "  G",
            "  ?",
            "  T",
            "  ?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            for expected, threshold in ("A", 0.4), ("R", 0.7), ("D", 0.95):
                self.assertEqual(
                    f"AC{expected}T",
                    consensusFromBAM(
                        bamFilename,
                        quiet=True,
                        referenceFasta=fastaFilename,
                        threshold=threshold,
                        noCoverage="reference",
                        ignoreQuality=self.ignoreQuality,
                    ).sequence,
                )

    def testGeneiousExamplesTie(self):
        """
        Test the tied counts example from
        https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
        """
        template = (
            "ACGT",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  A",
            "  ?",
            "  G",
            "  ?",
            "  G",
            "  ?",
            "  T",
            "  ?",
            "  T",
            "  ?",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            for expected, threshold in ("A", 0.4), ("D", 0.7), ("D", 0.95):
                self.assertEqual(
                    f"AC{expected}T",
                    consensusFromBAM(
                        bamFilename,
                        quiet=True,
                        referenceFasta=fastaFilename,
                        threshold=threshold,
                        noCoverage="reference",
                        ignoreQuality=self.ignoreQuality,
                    ).sequence,
                )

    def testTwoAgreeingSoftClipsNothingBefore(self):
        """
        Test that two soft-clipped regions that agree with each other are
        in the expected result when soft-clipped bases are included.
        """
        template = (
            "        TGATCTCC",
            "AGCCAGAATGATCTCC",
            "????????????????",
            "    AGAATGATCTCC",
            "    ????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "AGCCAGAATGATCTCC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoAgreeingSoftClipsNothingBeforeNoSoftClipped(self):
        """
        Test that two soft-clipped regions that agree with each other are
        ignored when soft-clipped bases are not including.
        """
        template = (
            "        TGATCTCC",
            "AGCCAGAATGATCTCC",
            "????????????????",
            "    AGAATGATCTCC",
            "    ????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TGATCTCC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoAgreeingInsertionsOneMatchingSomethingBefore(self):
        """
        Test that two insertions that agree with each other, with one also
        matching reference bases before the insertion give the expected result.
        """
        template = (
            "TT--------TGATCTCC",
            "TTAGCCAGAATGATCTCC",
            "??????????????????",
            "      AGAATGATCTCC",
            "      ????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TTAGCCAGAATGATCTCC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testTwoInsertions(self):
        """
        Test that two insertions that agree with each other, with one also
        matching reference bases before the insertion give the expected result.
        """
        template = (
            "TT--------TGAT--CTCC",
            "TTAGCCAGAATGA",
            "?????????????",
            "           GATGGCTCC",
            "           ?????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TTAGCCAGAATGATGGCTCC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214Insertion(self):
        """
        Test that an amino acid EPE sequence (here 'GAGCCAGAA') insertion
        into the SARS-CoV-2 spike nucleotide sequence at location 642 (amino
        acid location 214) works as expected.

        The nucleotide sequence below can be obtained via:

        $ ncbi-fetch-id.py MN908947.3 > MN908947.3.fasta
        $ describe-genome.py --feature S --printNtSeq < MN908947.3.fasta | \
              filter-fasta.py --keepSites 630-673 --quiet | tail -n 1

        I then inserted the 9 nucleotide sequence GAGCCAGAA before the TGAT...
        starting at position 642.
        """
        template = (
            "TAATTTAGTGCG---------TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
            "      AGTGCGGAGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTT",
            "      ??????????????????????????????????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TAATTTAGTGCGGAGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214InsertionRightSideExact(self):
        """
        Test that an amino acid EPE sequence (here 'GAGCCAGAA') insertion
        into the SARS-CoV-2 spike nucleotide sequence at location 642 (amino
        acid location 214) works as expected when the insertion is the very
        last part of the read, in which case it will be marked as a
        soft-clipped region.
        """
        template = (
            "TAATTTAGTGCG---------TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
            "      AGTGCGGAGCCAGAA",
            "      ???????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TAATTTAGTGCGGAGCCAGAATCAGGGTTTTTCGGCTTTAGAAC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214InsertionLeftSideExactSoftClipped(self):
        """
        Test that an amino acid EPE sequence (here 'GAGCCAGAA') insertion
        into the SARS-CoV-2 spike nucleotide sequence at location 642 (amino
        acid location 214) works as expected when the insertion sequence is
        the very beginning of the read, in which case it will be marked as a
        soft-clipped region.
        """
        template = (
            "TAATTTAGTGCG---------TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
            "            GAGCCAGAATGATCT",
            "            ???????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TAAGAGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214PartialInsertionInTwoReadsSoftClipped(self):
        """
        The amino acid EPE sequence (here 'GAGCCAGAA') insertion
        into the SARS-CoV-2 spike nucleotide sequence at location 642 (amino
        acid location 214) must work as expected, when the reads only
        partially cover it and are therefore both soft-clipped in that region.
        They overlap in a non-agreeing way and ambiguous nucleotide codes
        result.
        """
        template = (
            "TAATTTAGTGCG---------TGATCTCCCTCA",
            "      AGTGCGGAGCCA",
            "      ????????????",
            "             AGCCAGAATGATCTCCCTCA",
            "             ????????????????????",
        )

        # This example is a bit complicated. The reference and reads will
        # align as follows (with the '-' from the reference removed to make
        # it easier to see what's going on). The final line of hyphens shows
        # the ambiguous region.
        #
        # TAATTTAGTGCGTGATCTCCCTCA
        #       AGTGCGGAGCCA
        #     AGCCAGAATGATCTCCCTCA
        #       ------------

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TAATAGMSWGMRKRRYCWCCCTCA",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    includeSoftClipped=True,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214PartialInsertionInTwoReads(self):
        """
        This is the same as the immediately above test
        (testOmicronEPE214PartialInsertionInTwoReadsSoftClipped), but the
        reads are now fully aligned wtih the reference so there is no soft
        clipping. The alignment of the reads is the same and so is the result.
        """
        template = (
            "TAATTTAGTGCGTGATCTCCCTCA",
            "      AGTGCGGAGCCA",
            "      ????????????",
            "    AGCCAGAATGATCTCCCTCA",
            "    ????????????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TAATAGMSWGMRKRRYCWCCCTCA",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    insertionCountThreshold=1,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )

    def testOmicronEPE214PartialNoSoftClipped(self):
        """
        Test that a trailing part of the amino acid EPE sequence (here
        'AGCCAGAA') insertion into the SARS-CoV-2 spike nucleotide sequence
        that usually is found at location 642 (amino acid location 214) is
        excluded when it appears as a partial soft-clipped region and
        soft-clipped bases are not included.
        """
        template = (
            "        TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
            "AGCCAGAATGATCTCCCTCAGGGTTTTTCGGCTTT",
            "???????????????????????????????????",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "TGATCTCCCTCAGGGTTTTTCGGCTTTAGAAC",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )


@skipUnless(samtoolsInstalled(), "samtools is not installed")
class TestIgnoreQuality(TestCase, _Mixin):
    """
    Test making majority consensuses with quality ignored.
    """

    ignoreQuality = True


@skipUnless(samtoolsInstalled(), "samtools is not installed")
class TestWithQuality(TestCase, _Mixin):
    """
    Test making majority consensuses with quality considered. The expected
    return reslults for these tests differ from the results that would be
    obtained if quality was ignored, so for that reason these additional tests
    are not part of the common C{_Mixin} class.
    """

    ignoreQuality = False

    def testHighQualityDominates(self):
        """
        If one read has a very high quality base (here 'C', quality ']' = 60)
        that base should take precedence over two other reads that agree with
        each other but which have lower ('5' = 30) quality.
        """
        template = (
            "ACGT",
            "  A",
            "  5",
            "  A",
            "  5",
            "  C",
            "  ]",
        )

        with makeBAM(template) as (fastaFilename, bamFilename):
            self.assertEqual(
                "ACCT",
                consensusFromBAM(
                    bamFilename,
                    quiet=True,
                    referenceFasta=fastaFilename,
                    threshold=0.5,
                    noCoverage="reference",
                    ignoreQuality=self.ignoreQuality,
                ).sequence,
            )



#!/usr/bin/env python

"""
Given a JSON BLAST output file, a FASTA sequence file, and interesting
criteria, write a FASTA file (to sys.stdout) of the reads that match the
criteria.

Run with --help for help.
"""

import argparse
import sys

from dark.blast.alignments import BlastReadsAlignments
from dark.fasta import FastaReads
from dark.reads import Reads
from dark.titles import TitlesAlignments

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extract FASTA from matching BLAST hits.",
        epilog="Given JSON BLAST output files, a FASTA sequence file, "
        "and filtering criteria, produce a FASTA file on stdout "
        "containing the reads that match the criteria.",
    )

    # Args for the JSON BLAST and FASTA files.
    parser.add_argument(
        "json",
        metavar="BLAST-JSON-file",
        nargs="+",
        help="the JSON file of BLAST output.",
    )

    parser.add_argument(
        "--fasta",
        metavar="FASTA-file",
        required=True,
        help="the FASTA file of sequences that were given to BLAST.",
    )

    # Args for filtering on ReadsAlignments.
    parser.add_argument(
        "--minStart",
        type=int,
        default=None,
        help="Reads that start before this subject offset should not be shown.",
    )

    parser.add_argument(
        "--maxStop",
        type=int,
        default=None,
        help="Reads that end after this subject offset should not be shown.",
    )

    parser.add_argument(
        "--oneAlignmentPerRead",
        default=False,
        action="store_true",
        help="If C{True}, only keep the best alignment for each read.",
    )

    parser.add_argument(
        "--scoreCutoff",
        type=float,
        default=None,
        help=("A float score. Matches with scores worse than this will be ignored."),
    )

    parser.add_argument(
        "--maxHspsPerHit",
        type=int,
        default=None,
        help="A numeric max number of HSPs to show for each hit on hitId.",
    )

    parser.add_argument(
        "--whitelist",
        nargs="+",
        default=None,
        help="sequence titles that should be whitelisted",
    )

    parser.add_argument(
        "--blacklist",
        nargs="+",
        default=None,
        help="sequence titles that should be blacklisted",
    )

    parser.add_argument(
        "--titleRegex", default=None, help="a regex that sequence titles must match."
    )

    parser.add_argument(
        "--negativeTitleRegex",
        default=None,
        help="a regex that sequence titles must not match.",
    )

    parser.add_argument(
        "--truncateTitlesAfter",
        default=None,
        help="a string that titles will be truncated beyond. If a truncated "
        "title has already been seen, that title will be skipped.",
    )

    parser.add_argument(
        "--minSequenceLen",
        type=int,
        default=None,
        help="sequences of lesser length will be elided.",
    )

    parser.add_argument(
        "--maxSequenceLen",
        type=int,
        default=None,
        help="sequences of greater length will be elided.",
    )

    parser.add_argument(
        "--taxonomy",
        default=None,
        help="a string of the taxonomic group on which should be "
        'filtered. eg "Vira" will filter on viruses.',
    )

    # Args for filtering on TitlesAlignments.
    parser.add_argument(
        "--minMatchingReads",
        type=int,
        default=None,
        help="sequences that are matched by fewer reads will be elided.",
    )

    parser.add_argument(
        "--maxMatchingReads",
        type=int,
        default=None,
        help="sequences that are matched by more reads will be elided.",
    )

    parser.add_argument(
        "--minMedianScore",
        type=float,
        default=None,
        help="sequences that are matched with a median score that is "
        "worse will be elided.",
    )

    parser.add_argument(
        "--withScoreBetterThan",
        type=float,
        default=None,
        help="sequences that are matched without at least one score "
        "at least this good will be elided.",
    )

    parser.add_argument(
        "--minNewReads",
        type=float,
        default=None,
        help="The fraction of its reads by which a new read set must differ "
        "from all previously seen read sets in order to be considered "
        "acceptably different.",
    )

    args = parser.parse_args()
    reads = FastaReads(args.fasta)
    readsAlignments = BlastReadsAlignments(reads, args.json)

    # Convert white/blacklists lists to sets.
    if args.whitelist is not None:
        args.whitelist = set(args.whitelist)
    if args.blacklist is not None:
        args.blacklist = set(args.blacklist)

    readsAlignments.filter(
        minSequenceLen=args.minSequenceLen,
        maxSequenceLen=args.maxSequenceLen,
        minStart=args.minStart,
        maxStop=args.maxStop,
        oneAlignmentPerRead=args.oneAlignmentPerRead,
        maxHspsPerHit=args.maxHspsPerHit,
        scoreCutoff=args.scoreCutoff,
        whitelist=args.whitelist,
        blacklist=args.blacklist,
        titleRegex=args.titleRegex,
        negativeTitleRegex=args.negativeTitleRegex,
        truncateTitlesAfter=args.truncateTitlesAfter,
        taxonomy=args.taxonomy,
    )

    reads = Reads()
    count = 0

    if (
        args.minMatchingReads is None
        and args.maxMatchingReads is None
        and args.minMedianScore is None
        and args.withScoreBetterThan is None
        and args.minNewReads is None
    ):
        # No need to collect into titles, just get the read ids from
        # the matching alignments.
        for readAlignment in readsAlignments:
            reads.add(readAlignment.read)
            count += 1
    else:
        # We need to collect alignments into titles.
        titlesAlignments = TitlesAlignments(readsAlignments).filter(
            minMatchingReads=args.minMatchingReads,
            maxMatchingReads=args.maxMatchingReads,
            minMedianScore=args.minMedianScore,
            withScoreBetterThan=args.withScoreBetterThan,
            minNewReads=args.minNewReads,
        )

        for titleAlignments in titlesAlignments.values():
            for alignment in titleAlignments.alignments:
                reads.add(alignment.read)
                count += 1

    reads.save(sys.stdout)
    print("Found %d matching reads." % count, file=sys.stderr)



#!/usr/bin/env python

"""
Convert an e-value to a bit score.

Example usage:

$ e-value-to-bit-score.py --dbSize 168142520 --dbSequenceCount 5660 \
  --queryLength 111 --lengthAdjustment 26 --eValue 0.0813077725194
37.3537
"""

import argparse

from dark.blast.score import eValueToBitScore

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Convert a bit score to an e-value.")

    parser.add_argument(
        "--eValue", type=float, required=True, help="The e-value to convert."
    )

    parser.add_argument(
        "--dbSize",
        type=int,
        required=True,
        help="The total number of bases in the sequence database.",
    )

    parser.add_argument(
        "--dbSequenceCount",
        type=int,
        required=True,
        help="The number of sequences in the database.",
    )

    parser.add_argument(
        "--queryLength",
        type=int,
        required=True,
        help="The length of the query sequence.",
    )

    parser.add_argument(
        "--lengthAdjustment", type=int, required=True, help="The length adjustment."
    )

    args = parser.parse_args()
    print(
        eValueToBitScore(
            args.eValue,
            args.dbSize,
            args.dbSequenceCount,
            args.queryLength,
            args.lengthAdjustment,
        )
    )



#!/usr/bin/env python

"""
See the 'EPILOG' variable below, or (better) run with --help for help.
"""

import os
import sys

from dark.fasta import FastaReads

DEFAULT_BLAST_ARGS = ""
DEFAULT_BLAST_DB = "nt"
DEFAULT_BLAST_DB_DIR = "/usr/local/dark-matter/blast-dbs"
DEFAULT_EMAIL = "tcj25@cam.ac.uk"
DEFAULT_BLAST_EXECUTABLE_DIR = "/usr/local/dark-matter/blast/bin"
DEFAULT_BLAST_EXECUTABLE_NAME = "blastn"
DEFAULT_SEQUENCES_PER_BLAST = 100

EPILOG = """Given a FASTA file argument, write out the following:

  1) Files named 0.fasta, 1.fasta, 2.fasta, etc. each containing a maximum
     number of sequences (given by --seqs-per-blast).

  2) An HTCondor job spec file 'job.htcondor' that can be given to
     condor_submit to have it run BLAST on all the small FASTA files.

  3) A 'redo.sh' script that can be used to re-submit sub-FASTA files on which
     the initial processing failed.

  4) A 'post-process.sh' script that can be used to convert BLAST XML to our
     JSON format in the case that the last step of processing fails.

  5) A 'finalize.sh' script that will find all zero length .json result
     files and post-process them (if they have an XML file) and that also
     removes any empty .error files.

NOTE: files are overwritten. It is suggested you run this command in an
empty directory.
"""


def splitFASTA(params):
    """
    Read the FASTA file named params['fastaFile'] and print out its
    sequences into files named 0.fasta, 1.fasta, etc. with
    params['seqsPerJob'] sequences per file.
    """
    assert params["fastaFile"][-1] == "a", (
        "You must specify a file in fasta-format that ends in .fasta"
    )

    fileCount = count = seqCount = 0
    outfp = None
    with open(params["fastaFile"]) as infp:
        for seq in FastaReads(infp):
            seqCount += 1
            if count == params["seqsPerJob"]:
                assert outfp
                outfp.close()
                count = 0
            if count == 0:
                outfp = open("%d.fasta" % fileCount, "w")
                fileCount += 1
            count += 1
            assert outfp
            outfp.write(f">{seq.id}\n{seq.sequence}\n")
    assert outfp
    outfp.close()
    return fileCount, seqCount


def printJobSpec(params):
    """
    Write out a job spec file for HTCondor to process all the small
    FASTA input files via BLAST and our JSON post-processor.
    """
    with open("job.htcondor", "w") as outfp:
        outfp.write(
            """\
universe                  = vanilla
executable                = process.sh
should_transfer_files     = YES
when_to_transfer_output   = ON_EXIT
notify_user               = %(email)s
max_transfer_input_mb     = -1
max_transfer_output_mb    = -1
log                       = job.log

# Job summary:
#   FASTA input from %(fastaFile)s
#   %(sequenceCount)d sequences split into %(nJobs)d jobs of \
%(seqsPerJob)d sequences each.

arguments                 = $(Process) %(db)s %(blastArgs)s
input                     = $(Process).fasta
output                    = $(Process).done
error                     = $(Process).error
dont_encrypt_input_files  = $(Process).fasta
dont_encrypt_output_files = $(Process).json.bz2

queue %(nJobs)d
"""
            % params
        )


def printRedoScript(params):
    """
    Write out a shell script that writes a job spec file for HTCondor to
    process a single FASTA input file via BLAST and our JSON post-processor,
    runs that job, and removes the one-time spec file it wrote.
    """
    with open("redo.sh", "w") as outfp:
        outfp.write(
            """\
#!/bin/sh -e

case $# in
    0) echo "Usage: `basename $0` jobid1, jobid2, ..." >&2; exit 1;;
esac

tmp=redo.tmp.$$
trap "rm -f $tmp" 0 1 2 3 15

    cat >$tmp <<EOF
universe                  = vanilla
executable                = process.sh
should_transfer_files     = YES
when_to_transfer_output   = ON_EXIT
notify_user               = %(email)s
max_transfer_input_mb     = -1
max_transfer_output_mb    = -1
log                       = job.log
EOF

for jobid in "$@"
do
    cat >>$tmp <<EOF

arguments                 = $jobid %(db)s %(blastArgs)s
input                     = $jobid.fasta
output                    = $jobid.done
error                     = $jobid.error
dont_encrypt_input_files  = $jobid.fasta
dont_encrypt_output_files = $jobid.json.bz2
queue
EOF

    rm -f $jobid.json.bz2 $jobid.error $jobid.done
done

rm -f job.log

condor_submit $tmp
"""
            % params
        )

    # Make the script executable so we can run it.
    os.chmod("redo.sh", 0o755)


def printProcessScript(params):
    """
    Write out a simple process script to call BLAST and convert its XML to
    our compressed JSON format.
    """
    with open("process.sh", "w") as outfp:
        outfp.write(
            """\
#!/bin/sh

DM=/usr/local/dark-matter
export PYTHONPATH=$DM/dark-matter
export BLASTDB="%(dbDir)s"

jobid=$1
shift

db="$1"
shift

errs=$jobid.error

%(executableDir)s/%(executableName)s -query $jobid.fasta -db "$db" \
-outfmt 5 "$@" |
$DM/virtualenv/bin/python $DM/dark-matter/bin/convert-blast-xml-to-json.py \
--bzip2 > $jobid.json.bz2 2> $errs

if [ -s $errs ]
then
    echo "Completed WITH ERRORS ($errs) on `hostname` at `date`." > $jobid.done
else
    rm $errs
    echo "Completed on `hostname` at `date`." > $jobid.done
fi
"""
            % params
        )

    # Make the script executable so we can run it.
    os.chmod("process.sh", 0o755)


def printFinalizeScript(params):
    """
    Write out a script that post-processes any empty .json files that
    have non-empty .xml files, figures out which jobs should be re-run, and
    also removes zero-length error files.

    Note that we need bash in order to set the nullglob shell option. That
    prevents an error if there are no *.fasta files.
    """
    with open("finalize.sh", "w") as outfp:
        outfp.write(
            """\
#!/usr/bin/env bash

shopt -s nullglob

# redo will hold the numbers of jobs that need re-running, if any.
redo=

for i in *.fasta
do
    n=`echo $i | cut -f1 -d.`
    error=$n.error
    json=$n.json.bz2

    if [ -f $error ]
    then
        if [ -s $error ]
        then
            echo "WARNING: $error is non-empty." >&2
        else
            rm -f $error
        fi
    fi

    if [ -f $json ]
    then
        size=`wc -c < $json | awk '{print $1}'`
        # bzip2 run on empty input results in a file of 14 bytes.
        if [ $size -eq 14 ]
        then
            echo "WARNING: $json appears to be empty." >&2
        fi
    else
        echo "WARNING: $json does not exist. Job $n should be re-run." >&2
        redo="$redo $n"
    fi
done

if [ -n "$redo" ]
then
    echo "Some jobs must be re-run. Please execute the following:"
    echo "./redo.sh $redo"
fi
"""
            % params
        )

    # Make the script executable so we can run it.
    os.chmod("finalize.sh", 0o755)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Given a FASTA file, write an HTCondor job spec for BLAST",
        epilog=EPILOG,
    )
    parser.add_argument(
        "fasta", metavar="FASTA-file", help="the FASTA file of sequences to BLAST."
    )
    parser.add_argument(
        "--seqs-per-blast",
        metavar="N",
        type=int,
        default=DEFAULT_SEQUENCES_PER_BLAST,
        dest="seqsPerJob",
        help="the number (>0) of sequences to pass to BLAST on each run.",
    )
    parser.add_argument(
        "--blast-db-name",
        metavar="BLAST-database-name",
        default=DEFAULT_BLAST_DB,
        dest="db",
        help="the BLAST database to run against.",
    )
    parser.add_argument(
        "--email",
        metavar="name@host",
        default=DEFAULT_EMAIL,
        dest="email",
        help="the email address to send the job completed message to.",
    )
    parser.add_argument(
        "--blast-executable-name",
        metavar="BLAST-executable-name",
        default=DEFAULT_BLAST_EXECUTABLE_NAME,
        dest="executableName",
        choices=["blastn", "blastp", "blastx", "tblastn", "tblastx"],
        help="the name of the BLAST executable to run.",
    )
    parser.add_argument(
        "--blast-executable-dir",
        metavar="/path/to/BLAST/executables",
        default=DEFAULT_BLAST_EXECUTABLE_DIR,
        dest="executableDir",
        help="the directories that hold the BLAST executables.",
    )
    parser.add_argument(
        "--blast-db-dir",
        metavar="/BLAST/db/directory",
        default=DEFAULT_BLAST_DB_DIR,
        dest="dbDir",
        help="the directory where your BLAST database files live.",
    )
    parser.add_argument(
        "--blast-args",
        metavar="args...",
        default=DEFAULT_BLAST_ARGS,
        dest="blastArgs",
        help='additional arguments to pass to BLAST (e.g., "--task blastn".',
    )

    args = parser.parse_args()

    if args.seqsPerJob < 1:
        parser.print_help()
        sys.exit(1)

    params = {
        "blastArgs": args.blastArgs,
        "db": args.db,
        "dbDir": args.dbDir.rstrip("/"),
        "email": args.email,
        "executableName": args.executableName,
        "executableDir": args.executableDir.rstrip("/"),
        "fastaFile": args.fasta,
        "seqsPerJob": args.seqsPerJob,
    }
    params["nJobs"], params["sequenceCount"] = splitFASTA(params)
    printJobSpec(params)
    printProcessScript(params)
    printRedoScript(params)
    printFinalizeScript(params)

    print(
        (
            "%(sequenceCount)d sequences split into %(nJobs)d jobs of "
            "%(seqsPerJob)d sequences each." % params
        )
    )



import argparse
from pathlib import Path

from dark.extract_reference import extract_fasta, extract_sam
from dark.process import Executor


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=(
            "Extract a reference sequence from a FASTA file and its matches "
            "from a BAM file."
        ),
    )

    parser.add_argument(
        "--reference-id",
        required=True,
        metavar="ID",
        help="The ID of the reference sequence.",
    )

    parser.add_argument(
        "--out-dir",
        type=Path,
        required=True,
        metavar="DIR",
        help="The directory into which to put the output files.",
    )

    parser.add_argument(
        "--bam", "--sam", type=Path, required=True, help="The BAM or SAM file."
    )

    parser.add_argument("--fasta", type=Path, required=True, help="The FASTA file.")

    parser.add_argument("--quiet", action="store_true", help="Do not print a summary.")

    parser.add_argument(
        "--keep-description",
        action="store_true",
        help=(
            "Do not ignore the description (if any) of the sequence in the FASTA file. "
            "Normally, anything after the first space in the ID in the FASTA file will "
            "be ignored (for finding the ID in the FASTA file, in the output FASTA, "
            "and will not be included in the output BAM (where it is illegal). If you "
            "use this option, the description will not be ignored, so the "
            "--reference-id you give will be used in full. This may result in not "
            "finding what you are looking for (in the input FASTA)."
        ),
    )

    parser.add_argument(
        "--force", action="store_true", help="Overwrite pre-existing output files."
    )

    parser.add_argument(
        "--index-fasta",
        action="store_true",
        help="Index the extracted FASTA (using samtools faidx).",
    )

    parser.add_argument(
        "--keep-sam",
        action="store_true",
        help=(
            "Do not remove the intermediate SAM file with the matching records "
            "(produced en route to making a BAM file)"
        ),
    )

    return parser.parse_args()


def main() -> None:
    args = get_args()

    ref_id = args.reference_id
    out_dir = args.out_dir

    if not out_dir.exists():
        out_dir.mkdir()

    e = Executor()

    ref_len = extract_fasta(
        args.fasta,
        ref_id,
        out_dir,
        args.index_fasta,
        args.quiet,
        args.force,
        args.keep_description,
        e,
    )

    matches = extract_sam(
        args.bam, ref_id, ref_len, out_dir, args.keep_sam, args.force, e
    )

    if not args.quiet:
        s = "" if matches == 1 else "es"
        print(f"Found {matches} match{s} for {ref_id!r} in {str(args.bam)!r}.")


if __name__ == "__main__":
    main()
